"""The abstract syntax tree for representing a Schematron XML file.

Step one of loading a Schematron file is parsing it into the various `SchematronASTNode` instances.
At this step, `<include>` and `<extend href="">` tags are expanded and the referenced XML files are loaded and included.
Other referencing, like ID matching is done in a later stage.
"""
from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2022-10-06'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'GPL v3'

from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Self

from pyschematron.direct_mode.lib.ast import GenericASTNode


@dataclass(slots=True, frozen=True)
class SchematronASTNode(GenericASTNode):
    """Base class for all Schematron AST nodes.

    Since we aim for immutable objects, we do not allow the use of mutable objects in the nodes. Use either
    tuples (instead of lists) or frozendict instead of dict.
    """

    def __post_init__(self):
        for field_name, value in self.get_init_values().items():
            if isinstance(value, list):
                raise ValueError(f'No lists allowed for field "{field_name}", use a tuple instead.')

    def with_updated(self, **updated_items) -> Self:
        """Get a copy of this AST node with updated init values.

        This gets the current set of init values and updates those with the provided items.

        Args:
            updated_items: the keyword elements we wish to update

        Returns:
            A new copy of this node with the relevant items updated.
        """
        return type(self)(**(self.get_init_values() | updated_items))


@dataclass(slots=True, frozen=True)
class Schema(SchematronASTNode):
    """Representation of a `<schema>` tag.

    Although the standard only supports one `<diagnostics>` and `<properties>` element, we support multiple.

    Args:
        patterns: the listing of patterns
        namespaces: the listing of namespaces
        phases: the listing of phases
        diagnostics: the listing of diagnostics
        properties: the listing of properties
        paragraphs: the listing of paragraphs
        variables: the listing of variables
        title: the title of this scheme.
        default_phase: reference to the default phase of this Schema
        fpi: formal public identifier, a system-independent ID of this test
        icon: reference to a graphic file to be used in the error message
        id: the identifier of this element
        query_binding: the query binding for this scheme
        schema_version: an implementation defined version of the user's schema
        see: a URI or URL referencing background information
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    patterns: tuple[Pattern, ...] = tuple()
    namespaces: tuple[Namespace, ...] = tuple()
    phases: tuple[Phase, ...] = tuple()
    diagnostics: tuple[Diagnostics, ...] = tuple()
    properties: tuple[Properties, ...] = tuple()
    paragraphs: tuple[Paragraph, ...] = tuple()
    variables: tuple[Variable, ...] = tuple()
    title: Title | None = None

    default_phase: str | None = None
    fpi: str | None = None
    icon: str | None = None
    id: str | None = None
    query_binding: str | None = None
    schema_version: str | None = None
    see: str | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class Namespace(SchematronASTNode):
    """Representation of an `<ns>` tag.

    Args:
        prefix: the prefix for use in the Schematron and the Queries
        uri: the namespace's URI
    """
    prefix: str
    uri: str


@dataclass(slots=True, frozen=True)
class Phase(SchematronASTNode):
    """Representation of a `<phase>` tag.

    Phases define sets of patterns. When a phase is marked as active, only patterns within that phase definition are
    fired.

    Args:
        id: the identifier of this element
        active: listing of identifiers of active patterns
        variables: listing of variables scoped within this execution
        paragraphs: documentation paragraphs
        fpi: formal public identifier, a system-independent ID of this test
        icon: reference to a graphic file to be used in the error message
        see: a URI or URL referencing background information
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    id: str
    active: tuple[ActivePhase, ...] = tuple()
    variables: tuple[Variable, ...] = tuple()
    paragraphs: tuple[Paragraph, ...] = tuple()
    fpi: str | None = None
    icon: str | None = None
    see: str | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class ActivePhase(SchematronASTNode):
    """Representation of a `<active>` tag.

    Args:
        pattern_id: the ID of the pattern we should fire if this phase is active
        content: the text content, only used for documentation purposes.
    """
    pattern_id: str
    content: str | None = None


@dataclass(slots=True, frozen=True)
class Diagnostics(SchematronASTNode):
    """Representation of a `<diagnostics>` tag.

    Args:
        diagnostics: the listing of Diagnostic classes
    """
    diagnostics: tuple[Diagnostic, ...]


@dataclass(slots=True, frozen=True)
class Diagnostic(SchematronASTNode):
    """Representation of a `<diagnostic>` tag.

    Officially the `<name>` element is not allowed within a diagnostic. We allow it nonetheless.

    Args:
        content: the rich text content of this diagnostic.
        id: the identifier of this element
        fpi: formal public identifier, a system-independent ID of this test
        icon: reference to a graphic file to be used in the error message
        role: a description of the error message or the rule
        see: a URI or URL referencing background information
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    content: tuple[str | ValueOf | Name, ...]
    id: str
    fpi: str | None = None
    icon: str | None = None
    role: str | None = None
    see: str | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class Properties(SchematronASTNode):
    """Representation of a `<properties>` tag.

    Args:
        properties: the listing of Property classes
    """
    properties: tuple[Property, ...]


@dataclass(slots=True, frozen=True)
class Property(SchematronASTNode):
    """Representation of a `<property>` tag.

    Args:
        content: the rich text content of this property.
        id: the identifier of this element
        role: a description of the error message or the rule
        scheme: An IRI or public identifier specifying the notation used for the node's value.
    """
    content: tuple[str | ValueOf | Name, ...]
    id: str
    role: str | None = None
    scheme: str | None = None


@dataclass(slots=True, kw_only=True, frozen=True)
class Pattern(SchematronASTNode):
    """Abstract representation of a <pattern> tag.

    This can not be instantiated as is, one needs to load one of the subclasses for concrete or abstract patterns.

    Be reminded that the order of the rules matters. According to the Schematron definition, each node in an XML shall
    never be checked for multiple rules in one pattern. In the case of multiple matching rules, only the first matching
    rule is applied.

    Args:
        documents: xpath expression evaluating into one or more documents which must be checked instead of the
            current document.
        id: the identifier of this test
        fpi: formal public identifier, a system-independent ID of this test
        icon: reference to a graphic file to be used in the error message
        see: a URI or URL referencing background information
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    documents: XPathExpression | None = None
    id: str | None = None
    fpi: str | None = None
    icon: str | None = None
    see: str | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class ConcretePattern(Pattern):
    """A concrete pattern, this neither inherits another pattern, nor is an abstract pattern.

    Args:
        rules: the listing of rules
        variables: the listing of variables
        title: the title of this pattern
        paragraphs: the listing of paragraphs
    """
    rules: tuple[Rule, ...] = tuple()
    variables: tuple[Variable, ...] = tuple()
    paragraphs: tuple[Paragraph, ...] = tuple()
    title: Title | None = None


@dataclass(slots=True, frozen=True)
class AbstractPattern(Pattern):
    """An abstract pattern, coming from a pattern with the attribute `@abstract` set to True.

    Args:
        rules: the listing of rules
        variables: the listing of variables
        title: the title of this pattern
        paragraphs: the listing of paragraphs
    """
    rules: tuple[Rule, ...] = tuple()
    variables: tuple[Variable, ...] = tuple()
    paragraphs: tuple[Paragraph, ...] = tuple()
    title: Title | None = None


@dataclass(slots=True, frozen=True)
class InstancePattern(Pattern):
    """A pattern inheriting another pattern, i.e. patterns with the `is-a` attribute set.

    Args:
        params: the listing of pattern parameters
    """
    abstract_id_ref: str
    params: tuple[PatternParameter, ...] = tuple()


@dataclass(slots=True, frozen=True)
class PatternParameter(SchematronASTNode):
    """A parameter inside a pattern with the `is-a` attribute set.

    Args:
        name: the name of this parameter
        value: the value of this parameter
    """
    name: str
    value: str


@dataclass(slots=True, kw_only=True, frozen=True)
class Rule(SchematronASTNode):
    """Abstract representation of a <rule> tag.

    This can not be instantiated as is, rather instantiate one of the subclasses for concrete, abstract,
    or external rules.

    Args:
        checks: the listing of report and assert items in this rule
        variables: the listing of `<let>` variable declarations in this rule
        paragraphs: listing of paragraphs
        flag: name of the flag to which this test belongs, is set to True when this test is fired
        fpi: formal public identifier, a system-independent ID of this test
        icon: reference to a graphic file to be used in the error message
        role: a description of the error message or the rule
        see: a URI or URL referencing background information
        subject: a query referencing the node to which we assign an error message
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    checks: tuple[Check, ...] = tuple()
    variables: tuple[Variable, ...] = tuple()
    paragraphs: tuple[Paragraph, ...] = tuple()
    extends: tuple[Extends, ...] = tuple()
    flag: str | None = None
    fpi: str | None = None
    icon: str | None = None
    role: str | None = None
    see: str | None = None
    subject: XPathExpression | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class ConcreteRule(Rule):
    """Representation of a concrete <rule> tag (e.g. not abstract).

    Args:
        context: the attribute with the rule's context
        id: the identifier of this test (optional)
    """
    context: Query
    id: str | None = None


@dataclass(slots=True, frozen=True)
class AbstractRule(Rule):
    """Representation of an abstract <rule> tag.

    Args:
        id: the identifier of this test (required)
    """
    id: str


@dataclass(slots=True, frozen=True)
class ExternalRule(Rule):
    """Representation of an <rule> loaded from an external file using extends."""
    id: str | None = None


@dataclass(slots=True, frozen=True)
class Extends(SchematronASTNode):
    """Base class for <extends> tag representations used to extend a Rule with another rule."""


@dataclass(slots=True, frozen=True)
class ExtendsById(Extends):
    """Represents an <extends> tag which points to an abstract rule in this Schema.

    Args:
        id_ref: the identifier of a Rule inside the Schematron to which we refer
    """
    id_ref: str


@dataclass(slots=True, frozen=True)
class ExtendsExternal(Extends):
    """Represents an <extends> tag which has an AbstractRule loaded from another file.

    Args:
        rule: an ExternalRule loaded from another file.
        file_path: the path from which the rule was loaded
    """
    rule: ExternalRule
    file_path: Path


@dataclass(slots=True, frozen=True)
class Check(SchematronASTNode):
    """Base class for `<assert>` and `<report>` elements.

    Args:
        test: the attribute with the test condition
        content: the mixed text content, can contain `<emph>`, `<span>`, `<dir>`,
            `<value-of>`, and `<name>` as flat text.
        diagnostics: listing of IDs referencing a diagnostic elements
        flag: name of the flag to which this test belongs, is set to True when this test is fired
        fpi: formal public identifier, a system-independent ID of this test
        icon: reference to a graphic file to be used in the error message
        id: the identifier of this test
        properties: listing of identifiers of property items
        role: a description of the error message or the rule
        see: a URI or URL referencing background information
        subject: an XPath expression referencing the node to which we assign an error message
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    test: Query
    content: tuple[str | ValueOf | Name, ...]
    diagnostics: tuple[str, ...] | None = None
    flag: str | None = None
    fpi: str | None = None
    icon: str | None = None
    id: str | None = None
    properties: tuple[str, ...] | None = None
    role: str | None = None
    see: str | None = None
    subject: XPathExpression | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class Assert(Check):
    """Representation of an `<assert>` tag."""


@dataclass(slots=True, frozen=True)
class Report(Check):
    """Representation of a `<report>` tag."""


@dataclass(slots=True, frozen=True)
class Variable(SchematronASTNode):
    """Abstract representation of a `<let>` tag.

    Let tags can be defined in two ways, one like `<let name="..." value="..."/>` and another like
    `<let name="..."><some-xml xmlns="...">...</some-xml></let>`. That is, in the second case the value of the
    variable is a separate XML tree.

    Args:
        name: the name attribute
    """
    name: str


@dataclass(slots=True, frozen=True)
class QueryVariable(Variable):
    """Representation of a `<let>` tag with a Query attribute.

    Args:
        value: the value attribute
    """
    value: Query


@dataclass(slots=True, frozen=True)
class XMLVariable(Variable):
    """Representation of a `<let>` tag with the value loaded from the node's content.

    In this case the `<let>` looks like: `<let name="..."><some-xml xmlns="...">...</some-xml></let>` and the
    value would be `<some-xml xmlns="...">...</some-xml>`, i.e. some XML in some namespace not Schematron's.

    Args:
        value: the content of the `<let>` element.
    """
    value: str


@dataclass(slots=True, frozen=True)
class Paragraph(SchematronASTNode):
    """Representation of a `<p>` tag.

    Although the Schematron ISO standard does not prescribe the `xml:lang` and `xml:space` attribute for paragraphs, we
    add them nonetheless.

    Args:
        content: the text content of this paragraph
        class_: the class attribute
        icon: the icon attribute
        id: the identifier
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    content: str
    class_: str | None = None
    icon: str | None = None
    id: str | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None


@dataclass(slots=True, frozen=True)
class Title(SchematronASTNode):
    """Representation of a `<title>` tag.

    Args:
        content: the text content of this title
    """
    content: str


@dataclass(slots=True, frozen=True)
class Query(SchematronASTNode):
    """Representation of a Query used in the Schematron AST nodes."""
    query: str


@dataclass(slots=True, frozen=True)
class XPathExpression(SchematronASTNode):
    """Representation of an XPath expression.

    Since Schematron queries are abstract entities defined by the query binding in the Schematron root, we also need
    an XPath expression object for elements which may only be XPath expressions.
    """
    expression: str


@dataclass(slots=True, frozen=True)
class RichTextContent(SchematronASTNode):
    """Specific subclass for rich text content.

    Rich text in Schematron allows for `<dir>`, `<emph>`, `<span>`, `<value-of>` and sometimes `<name>` elements.
    In this package we allow `<name>` in all rich text content.

    The non-Schematron specific elements dir, emph and span are handled as text content by the parser. The name and
    value-of are specified by the AST nodes.

    By subclassing these nodes we allow to group them.
    """


@dataclass(slots=True, frozen=True)
class ValueOf(RichTextContent):
    """Representation of a `<value-of>` node."""
    select: Query


@dataclass(slots=True, frozen=True)
class Name(RichTextContent):
    """Representation of a `<name>` node."""
    path: Query | None = None



"""This module defines abstract base types for an Abstract Syntax Tree (AST) and an AST visitor."""
from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2024-03-21'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from abc import abstractmethod, ABCMeta
from dataclasses import dataclass, fields
from typing import Any, Mapping, Iterable, Self


@dataclass(slots=True, frozen=True)
class GenericASTNode:
    """Abstract base class for Abstract Syntax Tree (AST) nodes.

    Each node in an AST (also the root) is of this type. Since we, in general, aim for immutability, the AST nodes
    are defined as frozen dataclasses with slots. If you want true immutability, avoid dictionaries and lists in
    your AST implementations.

    This class already implements the visitor pattern using the :class:`GenericASTVisitor`.
    """

    def accept_visitor(self, visitor: GenericASTVisitor) -> Any:
        """Accept a visitor on this node.

        Since Python allows polymorphic return values, we allow the visitor pattern to return values.

        Args:
            visitor: the visitor we accept and call

        Returns:
            The result of the visitor.
        """
        return visitor.visit(self)

    def get_init_values(self) -> dict[str, Any]:
        """Get the initialisation values with which this class was instantiated.

        Returns:
            A dictionary with the arguments which instantiated this object.
        """
        return {field.name: getattr(self, field.name) for field in fields(self)}

    def get_children(self) -> list[Self]:
        """Get a list of all the AST nodes in this node.

        This should return all the references to AST nodes in this node, all bundled in one list.

        Returns:
            All the child nodes in this node
        """
        def get_ast_nodes(init_values):
            children = []

            if isinstance(init_values, GenericASTNode):
                children.append(init_values)
            elif isinstance(init_values, Mapping):
                for el in init_values.values():
                    children.extend(get_ast_nodes(el))
            elif isinstance(init_values, Iterable) and not isinstance(init_values, str):
                for el in init_values:
                    children.extend(get_ast_nodes(el))

            return children

        return get_ast_nodes(self.get_init_values())


class GenericASTVisitor[T: GenericASTNode](metaclass=ABCMeta):
    """Generic base class for visitors, according to the visitor design pattern.

    Instead of a typed double dispatch we use dynamic double dispatching in which each node, when visited, calls
    the :meth:``visit` of this class instead of a visit method for each node type. This makes it easier to
    do edits on class names since the types can be looked up by an IDE.

    We use the generic type hint `T` to ensure solid type hinting for implementing visitors.
    """

    @abstractmethod
    def visit(self, ast_node: T) -> Any:
        """Visit the AST node.

        This uses dynamic dispatch to accept all types of AST nodes.

        Since Python allows polymorphic return values, we allow the visitor pattern to return values.

        Args:
            ast_node: an AST node of any type

        Returns:
            The result of the visitor.
        """

    def apply(self, ast_node: T) -> Any:
        """Convenience method to apply this visitor on the indicated node and get the result value.

        Args:
            ast_node: the node on which to apply this visitor

        Returns:
            The result value from :meth:`get_result`
        """
        return ast_node.accept_visitor(self)




__author__ = 'Robbert Harms'
__date__ = '2023-02-18'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

from abc import ABCMeta, abstractmethod
from typing import override

from pyschematron.direct_mode.schematron.ast import (SchematronASTNode, Check, Variable, Paragraph, Extends,
                                                     ConcreteRule, ExternalRule, AbstractRule, Query, Rule,
                                                     ConcretePattern, Pattern, Namespace, Schema, Title,
                                                     AbstractPattern, InstancePattern, PatternParameter, Phase,
                                                     ActivePhase, Diagnostics, Properties, XPathExpression)
from pyschematron.direct_mode.schematron.parsers.xml.utils import parse_attributes


class SchematronASTNodeBuilder(metaclass=ABCMeta):
    """Builder pattern for delayed construction of Schematron nodes."""

    @abstractmethod
    def build(self) -> SchematronASTNode:
        """Build a Schematron node based on the information in this builder.

        Returns:
            The constructed schematron element.
        """


class RuleBuilder(SchematronASTNodeBuilder, metaclass=ABCMeta):

    def __init__(self):
        """Construct a Rule node out of the parts provided.

        Can not be used directly, one needs to use one of the specialized subclasses.
        """
        self.checks: list[Check] = []
        self.variables: list[Variable] = []
        self.paragraphs: list[Paragraph] = []
        self.extends: list[Extends] = []
        self.attributes = {}

    def add_checks(self, nodes: list[Check]):
        """Add a list of Check nodes (Report or Assert).

        Args:
            nodes: the nodes to add to the list of checks.
        """
        self.checks.extend(nodes)

    def add_variables(self, nodes: list[Variable]):
        """Add a list of Variable nodes.

        Args:
            nodes: the nodes to add to the list of variables.
        """
        self.variables.extend(nodes)

    def add_paragraphs(self, nodes: list[Paragraph]):
        """Add a list of Paragraph nodes.

        Args:
            nodes: the nodes to add to the list of paragraphs.
        """
        self.paragraphs.extend(nodes)

    def add_extends(self, nodes: list[Extends]):
        """Add a list of extends nodes.

        Args:
            nodes: the nodes to add to the list of extends
        """
        self.extends.extend(nodes)

    def add_attributes(self, element_attributes: dict[str, str]):
        """Add all the attributes of the XML Rule element in one go.

        Args:
            element_attributes: dictionary of attributes taken from the XML node
        """
        allowed_attributes = ['context', 'subject', 'flag', 'fpi', 'icon', 'id', 'role', 'see',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            'context': lambda k, v: {k: Query(v)},
            'subject': lambda k, v: {k: XPathExpression(v)},
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element_attributes, allowed_attributes, attribute_handlers)
        self.attributes.update(attributes)


class ConcreteRuleBuilder(RuleBuilder):

    @override
    def build(self) -> ConcreteRule:
        if 'context' not in self.attributes:
            raise ValueError('A concrete rule must have a context.')

        return ConcreteRule(checks=tuple(self.checks), variables=tuple(self.variables),
                            paragraphs=tuple(self.paragraphs), extends=tuple(self.extends), **self.attributes)


class AbstractRuleBuilder(RuleBuilder):

    @override
    def build(self) -> AbstractRule:
        if 'context' in self.attributes:
            raise ValueError('An abstract rule can not have a context.')

        if 'id' not in self.attributes:
            raise ValueError('An abstract rule must have an id.')

        return AbstractRule(checks=tuple(self.checks), variables=tuple(self.variables),
                            paragraphs=tuple(self.paragraphs), extends=tuple(self.extends), **self.attributes)


class ExternalRuleBuilder(RuleBuilder):

    @override
    def build(self) -> ExternalRule:
        if 'context' in self.attributes:
            raise ValueError('An external rule can not have a context.')

        return ExternalRule(checks=tuple(self.checks), variables=tuple(self.variables),
                            paragraphs=tuple(self.paragraphs), extends=tuple(self.extends), **self.attributes)


class PatternBuilder(SchematronASTNodeBuilder, metaclass=ABCMeta):

    def __init__(self):
        """Construct a Pattern node out of the parts provided.

        Can not be used directly, one needs to use one of the specialized subclasses.
        """
        self.rules: list[Rule] = []
        self.variables: list[Variable] = []
        self.title: Title | None = None
        self.paragraphs: list[Paragraph] = []
        self.pattern_parameters: list[PatternParameter] = []
        self.attributes = {}

    def add_rules(self, nodes: list[Rule]):
        """Add a list of Rule nodes

        Args:
            nodes: the nodes to add to the list of checks.
        """
        self.rules.extend(nodes)

    def add_variables(self, nodes: list[Variable]):
        """Add a list of Variable nodes.

        Args:
            nodes: the nodes to add to the list of variables.
        """
        self.variables.extend(nodes)

    def set_title(self, node: Title | None):
        """Set the title node.

        Args:
            node: the title node
        """
        self.title = node

    def add_paragraphs(self, nodes: list[Paragraph]):
        """Add a list of Paragraph nodes.

        Args:
            nodes: the nodes to add to the list of paragraphs.
        """
        self.paragraphs.extend(nodes)

    def add_parameters(self, nodes: list[PatternParameter]):
        """Add a list of PatternParameter nodes.

        Args:
            nodes: the nodes to add to the list of parameters.
        """
        self.pattern_parameters.extend(nodes)

    def add_attributes(self, element_attributes: dict[str, str]):
        """Add all the attributes of the XML Pattern element in one go.

        Args:
            element_attributes: dictionary of attributes taken from the XML node
        """
        allowed_attributes = ['documents', 'fpi', 'icon', 'id', 'see', 'is-a',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            'documents': lambda k, v: {k: Query(v)},
            'is-a': lambda k, v: {'abstract_id_ref': v},
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element_attributes, allowed_attributes, attribute_handlers)
        self.attributes.update(attributes)


class ConcretePatternBuilder(PatternBuilder):

    @override
    def build(self) -> ConcretePattern:
        return ConcretePattern(rules=tuple(self.rules), variables=tuple(self.variables),
                               paragraphs=tuple(self.paragraphs), title=self.title, **self.attributes)


class AbstractPatternBuilder(PatternBuilder):

    @override
    def build(self) -> AbstractPattern:
        return AbstractPattern(rules=tuple(self.rules), variables=tuple(self.variables),
                               paragraphs=tuple(self.paragraphs), title=self.title, **self.attributes)


class InstancePatternBuilder(PatternBuilder):

    @override
    def build(self) -> InstancePattern:
        return InstancePattern(params=tuple(self.pattern_parameters), **self.attributes)


class PhaseBuilder(SchematronASTNodeBuilder):

    def __init__(self):
        """Construct a Phase node out of the parts provided."""
        self.active: list[ActivePhase] = []
        self.variables: list[Variable] = []
        self.paragraphs: list[Paragraph] = []
        self.attributes = {}

    @override
    def build(self) -> Phase:
        return Phase(active=tuple(self.active), variables=tuple(self.variables),
                     paragraphs=tuple(self.paragraphs), **self.attributes)

    def add_active(self, nodes: list[ActivePhase]):
        """Add a list of ActivePhase nodes

        Args:
            nodes: the nodes to add to the list of active phases.
        """
        self.active.extend(nodes)

    def add_variables(self, nodes: list[Variable]):
        """Add a list of Variable nodes.

        Args:
            nodes: the nodes to add to the list of variables.
        """
        self.variables.extend(nodes)

    def add_paragraphs(self, nodes: list[Paragraph]):
        """Add a list of Paragraph nodes.

        Args:
            nodes: the nodes to add to the list of paragraphs.
        """
        self.paragraphs.extend(nodes)

    def add_attributes(self, element_attributes: dict[str, str]):
        """Add all the attributes of the XML Pattern element in one go.

        Args:
            element_attributes: dictionary of attributes taken from the XML node
        """
        allowed_attributes = ['fpi', 'icon', 'id', 'see',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element_attributes, allowed_attributes, attribute_handlers)
        self.attributes.update(attributes)


class SchemaBuilder(SchematronASTNodeBuilder):

    def __init__(self):
        """Construct a Schema node out of the parts provided."""
        self.patterns: list[Pattern] = []
        self.namespaces: list[Namespace] = []
        self.diagnostics: list[Diagnostics] = []
        self.properties: list[Properties] = []
        self.title: Title | None = None
        self.variables: list[Variable] = []
        self.paragraphs: list[Paragraph] = []
        self.phases: list[Phase] = []
        self.attributes = {}

    @override
    def build(self) -> Schema:
        return Schema(patterns=tuple(self.patterns), namespaces=tuple(self.namespaces), phases=tuple(self.phases),
                      paragraphs=tuple(self.paragraphs), variables=tuple(self.variables),
                      diagnostics=tuple(self.diagnostics), properties=tuple(self.properties), title=self.title,
                      **self.attributes)

    def add_patterns(self, nodes: list[Pattern]):
        """Add a list of Pattern nodes

        Args:
            nodes: the nodes to add to the list of patterns.
        """
        self.patterns.extend(nodes)

    def add_namespaces(self, nodes: list[Namespace]):
        """Add a list of Namespace nodes

        Args:
            nodes: the nodes to add to the list of namespaces.
        """
        self.namespaces.extend(nodes)

    def add_phases(self, nodes: list[Phase]):
        """Add a list of Phase nodes

        Args:
            nodes: the nodes to add to the list of phases.
        """
        self.phases.extend(nodes)

    def add_diagnostics(self, nodes: list[Diagnostics]):
        """Add a list of Diagnostics nodes

        Args:
            nodes: the nodes to add to the list of diagnostics.
        """
        self.diagnostics.extend(nodes)

    def add_properties(self, nodes: list[Properties]):
        """Add a list of Properties nodes

        Args:
            nodes: the nodes to add to the list of properties.
        """
        self.properties.extend(nodes)

    def add_variables(self, nodes: list[Variable]):
        """Add a list of Variable nodes.

        Args:
            nodes: the nodes to add to the list of variables.
        """
        self.variables.extend(nodes)

    def add_paragraphs(self, nodes: list[Paragraph]):
        """Add a list of Paragraph nodes.

        Args:
            nodes: the nodes to add to the list of paragraphs.
        """
        self.paragraphs.extend(nodes)

    def set_title(self, node: Title | None):
        """Set the title node.

        Args:
            node: the title node
        """
        self.title = node

    def add_attributes(self, element_attributes: dict[str, str]):
        """Add all the attributes of the XML Schema element in one go.

        Args:
            element_attributes: dictionary of attributes taken from the XML node
        """
        allowed_attributes = ['defaultPhase', 'fpi', 'icon', 'id',
                              'queryBinding', 'schemaVersion', 'see',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            'defaultPhase': lambda k, v: {'default_phase': v},
            'queryBinding': lambda k, v: {'query_binding': v},
            'schemaVersion': lambda k, v: {'schema_version': v},
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element_attributes, allowed_attributes, attribute_handlers)
        self.attributes.update(attributes)



from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2023-05-06'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'GPL v3'

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

from lxml.etree import _ElementTree

from pyschematron.direct_mode.schematron.ast import ConcreteRule, Assert, Report, ConcretePattern, Schema
from pyschematron.direct_mode.xml_validation.queries.base import EvaluationContext
from pyschematron.direct_mode.xml_validation.results.xml_nodes import XMLNode


@dataclass(slots=True, frozen=True)
class ValidationResult:
    """Type class for the validation results. """


@dataclass(slots=True, frozen=True)
class XMLDocumentValidationResult(ValidationResult):
    """Result class for the full evaluation of the entire XML document.

    This encapsulates the processing of all patterns over all nodes.

    Args:
        xml_information: the knowledge of the XML
        schema_information: information about the applied Schema
        node_results: the results over all nodes
    """
    xml_information: XMLInformation
    schema_information: SchemaInformation
    node_results: tuple[FullNodeResult, ...]

    def is_valid(self) -> bool:
        """Return True if the XML document was considered valid, False otherwise.

        According to the specifications, a successful report is considered a failure. As such, this method considers
        an XML document to be valid if none of the assertions and none of the reports were raised.

        Returns:
            True if the document passed the Schematron validation, False otherwise.
        """
        for node_result in self.node_results:
            if not node_result.is_valid():
                return False
        return True


@dataclass(slots=True, frozen=True)
class XMLInformation(ValidationResult):
    """Container for the knowledge of the XML being validated.

    This encapsulates the processing of all patterns over all nodes.

    Args:
        xml_document: the XML document provided as input
    """
    xml_document: _ElementTree


@dataclass(slots=True, frozen=True)
class SchemaInformation(ValidationResult):
    """Container for the information of the Schematron used during validation.

    Args:
        schema: the Schema AST node used during evaluation
        phase: the phase used in evaluation
        schematron_base_path: the base path from which we loaded the Schematron file, provided for context.
    """
    schema: Schema
    phase: str | Literal['#ALL', '#DEFAULT'] | None = None,
    schematron_base_path: Path | None = None


@dataclass(slots=True, frozen=True)
class BaseXMLNodeResult(ValidationResult):
    """Base class for the result of processing a specific XML node.

    Args:
        xml_node: the node on which we are reporting the processing result.
        evaluation_context: the context in which the node was processed.
            This should not be specialized to the context in which the node was processed. For example, for a processed
            pattern, this should be the "outside" evaluation context without the parameters inside the pattern.
    """
    xml_node: XMLNode
    evaluation_context: EvaluationContext


@dataclass(slots=True, frozen=True)
class FullNodeResult(BaseXMLNodeResult):
    """Result class for the full evaluation of an XML node.

    This encapsulates the processing of all the patterns over the indicated XML node.

    Args:
        pattern_results: the results of all the patterns
    """
    pattern_results: tuple[PatternResult, ...]

    def is_valid(self) -> bool:
        """Return True if all patterns yielded a valid results, False otherwise.

        Returns:
            True if the document passed the Schematron validation, False otherwise.
        """
        for pattern_result in self.pattern_results:
            if not pattern_result.is_valid():
                return False
        return True


@dataclass(slots=True, frozen=True)
class PatternResult(BaseXMLNodeResult):
    """Result class for evaluating a pattern on a node.

    Args:
        pattern: a reference to the evaluated pattern
        rule_results: a list of the rule results for each rule in the pattern.
    """
    pattern: ConcretePattern
    rule_results: tuple[RuleResult, ...]

    def has_fired_rule(self) -> bool:
        """Check if this pattern result has a fired rule or not.

        Returns:
            True if there was an active rule in this pattern for the node, False otherwise
        """
        return any(result.is_fired() for result in self.rule_results)

    def is_valid(self) -> bool:
        """Return True if all rules yielded a valid results, False otherwise.

        Returns:
            True if the document passed the Schematron validation, False otherwise.
        """
        for rule_result in self.rule_results:
            if isinstance(rule_result, FiredRuleResult):
                if not rule_result.is_valid():
                    return False
        return True


@dataclass(slots=True, frozen=True)
class RuleResult(BaseXMLNodeResult, metaclass=ABCMeta):
    """Base class for skipped, fired, and suppressed rules.

    Since we process all rules we need a way to indicate if a rule was skipped, fired, or suppressed.
    This base class creates a base type for the different rule results.

    Args:
        rule: the rule which was processed
    """
    rule: ConcreteRule

    @abstractmethod
    def is_skipped(self) -> bool:
        """Check if this rule was skipped or not.

        Returns:
            True if the rule was skipped, False otherwise
        """

    @abstractmethod
    def is_fired(self) -> bool:
        """Check if this rule was fired or not.

        Returns:
            True if the rule was fired, False otherwise
        """

    @abstractmethod
    def is_suppressed(self) -> bool:
        """Check if this rule was suppressed or not.

        Returns:
            True if the rule was suppressed, False otherwise
        """


@dataclass(slots=True, frozen=True)
class SkippedRuleResult(RuleResult):
    """Indicates the result of a rule which was skipped because the context did not match."""

    def is_skipped(self) -> bool:
        return True

    def is_fired(self) -> bool:
        return False

    def is_suppressed(self) -> bool:
        return False


@dataclass(slots=True, frozen=True)
class SuppressedRuleResult(RuleResult):
    """Indicates the result of a rule which was shadowed by a preceding rule."""

    @classmethod
    def from_fired_rule_result(cls, fired_rule_result: FiredRuleResult):
        """Generated a suppressed rule result from the result of a fired rule.

        This is a convenience method to turn a fired rule in a suppressed rule.

        Args:
            fired_rule_result: the fired result we would like to transform
        """
        return cls(fired_rule_result.xml_node, fired_rule_result.evaluation_context, fired_rule_result.rule)

    def is_skipped(self) -> bool:
        return False

    def is_fired(self) -> bool:
        return False

    def is_suppressed(self) -> bool:
        return True


@dataclass(slots=True, frozen=True)
class FiredRuleResult(RuleResult):
    """The result of checking the asserts and reports of a Rule on an XML node.

    Args:
        check_results: the results of the checks
        subject_node: the node referenced by the subject attribute of the Schematron rule.
    """
    check_results: list[CheckResult]
    subject_node: XMLNode | None

    def is_skipped(self) -> bool:
        return False

    def is_fired(self) -> bool:
        return True

    def is_suppressed(self) -> bool:
        return False

    def is_valid(self) -> bool:
        """Return True if all checks yielded a valid results, False otherwise.

        Returns:
            True if the document passed the Schematron validation, False otherwise.
        """
        for check_result in self.check_results:
            if check_result.check_result:
                return False
        return True


@dataclass(slots=True, frozen=True)
class CheckResult(BaseXMLNodeResult):
    """The result of checking a Schematron assert or report on an XML node.

    The test result stored in this class represents if the test in the check was true or false. As such,
    it is independent on the nature of the check. A false test result for an assertion means a failure, which will be
    reported, while only a true test result for a report is reported. If you want this derived message, use the
    dynamic check result property.

    Args:
        check: the check which was run
        test_result: the result of the test in the check.
        text: the text result from the rich text content.
        subject_node: the node referenced by the subject attribute of the Schematron check.
    """
    check: Assert | Report
    test_result: bool
    text: str
    subject_node: XMLNode | None
    property_results: tuple[PropertyResult, ...] | None = None
    diagnostic_results: tuple[DiagnosticResult, ...] | None = None

    @property
    def check_result(self) -> bool:
        """Get the result of the check.

        In Schematron, tests can be written in one of two ways:

            <sch:assert> outputs a message if an XPath test evaluates to false.
            <sch:report> outputs a message if an XPath test evaluates to true.

        The test result stored in this class represents the state of the test result, not the final outcome
        of the check. For that, there is this method.

        This checks if the result was a pass or not, it returns a value based on the following combinations:

        +--------+-------------+--------------+
        |  Check | Test result | Return value |
        +========+=============+==============+
        | Assert | true        | false        |
        | Assert | false       | true         |
        | Report | true        | true         |
        | Report | false       | false        |
        +--------+-------------+--------------+

        Returns:
            If the return value is true, we are either dealing with a failed assert, or a successful report.
            If the return value is false, we have a successful assert, or a failed report.
        """
        if isinstance(self.check, Assert):
            return not self.test_result
        else:
            return self.test_result


@dataclass(slots=True, frozen=True)
class PropertyResult(ValidationResult):
    """Result of evaluating a property indicated by a check.

    Args:
        text: resulting text
        property_id: identifier of this property
        role: the role attribute for this property, copied from the properties' role attribute
        scheme: the scheme attribute for this property, copied from the properties' scheme attribute
    """
    text: str
    property_id: str
    role: str | None = None
    scheme: str | None = None


@dataclass(slots=True, frozen=True)
class DiagnosticResult(ValidationResult):
    """Result of evaluating a diagnostic indicated by a check.

    Args:
        text: resulting text
        diagnostic_id: identifier of this diagnostic
        xml_lang: the xml language attribute for this diagnostic
        xml_space: the xml_space attribute from the diagnostic
    """
    text: str
    diagnostic_id: str
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None



__author__ = 'Robbert Harms'
__date__ = '2024-03-17'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from abc import ABCMeta, abstractmethod
from typing import override

from lxml.etree import Element, _Element
from pyschematron.direct_mode.svrl.ast import SchematronOutput, SVRLNode, NSPrefixInAttributeValues, Text, \
    ActivePattern, MetaData, FiredRule, FailedAssert, SuppressedRule, SuccessfulReport, PropertyReference, \
    DiagnosticReference
from pyschematron.direct_mode.svrl.svrl_visitors import SVRLASTVisitor


class SVRLWriter(metaclass=ABCMeta):
    """Type class for SVRL XML writers."""

    @abstractmethod
    def create_xml(self, schematron_output: SchematronOutput) -> _Element:
        """Transform a Schematron output SVRL root node, to an XML element.

        Args:
            schematron_output: the root node of an SVRL AST.

        Returns:
            An XML representation of the provided SVRL output node.
        """


class LxmlSVRLWriter(SVRLWriter):
    """SVRL writer using the Lxml library."""

    @override
    def create_xml(self, schematron_output: SchematronOutput) -> _Element:
        nsmap = {
            'svrl': 'http://purl.oclc.org/dsdl/svrl',
            'sch': 'http://purl.oclc.org/dsdl/schematron',
            'xs': 'http://www.w3.org/2001/XMLSchema'
        }
        writer_visitor = _SVRLWriterVisitor(nsmap)
        return writer_visitor.visit(schematron_output)


class _SVRLWriterVisitor(SVRLASTVisitor):

    def __init__(self, nsmap: dict[str, str]):
        """SVRL XML writer using the visitor pattern.

        Args:
            nsmap: the default namespaces to apply in the SVRL.
        """
        self._nsmap = nsmap

    @override
    def visit(self, svrl_node: SVRLNode) -> _Element | None:
        match svrl_node:
            case SchematronOutput():
                return self._process_schematron_output(svrl_node)
            case NSPrefixInAttributeValues():
                return self._process_ns_prefix_node(svrl_node)
            case MetaData():
                return self._process_metadata(svrl_node)
            case ActivePattern():
                return self._process_active_pattern_node(svrl_node)
            case FiredRule():
                return self._process_fired_rule_node(svrl_node)
            case SuppressedRule():
                return self._process_suppressed_rule_node(svrl_node)
            case FailedAssert():
                return self._process_failed_assert(svrl_node)
            case SuccessfulReport():
                return self._process_successful_report(svrl_node)
            case Text():
                return self._process_text_node(svrl_node)
            case PropertyReference():
                return self._process_property_reference(svrl_node)
            case DiagnosticReference():
                return self._process_diagnostic_reference(svrl_node)
        return None

    def _process_schematron_output(self, schematron_output: SchematronOutput) -> _Element:
        """Process the SchematronOutput node, the root of the SVRL.

        Args:
            schematron_output: the SVRL root node

        Returns:
            An element representing the SVRL report.
        """
        for ns_prefix in schematron_output.ns_prefix_in_attribute_values:
            if ns_prefix.prefix in self._nsmap:
                raise ValueError(f'Provided namespace "{ns_prefix.prefix}" is not allowed in the Schematron, '
                                 f'it overlaps with the default SVRL namespaces.')
            else:
                self._nsmap[ns_prefix.prefix] = ns_prefix.uri

        node_attributes = {
            'phase': schematron_output.phase,
            'schemaVersion': schematron_output.schema_version,
            'title': schematron_output.title
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        root = Element(f'{{{self._nsmap["svrl"]}}}schematron-output', attrib=final_attributes, nsmap=self._nsmap)

        for text_node in schematron_output.texts:
            root.append(self.visit(text_node))

        for ns_prefix in schematron_output.ns_prefix_in_attribute_values:
            root.append(self.visit(ns_prefix))

        if schematron_output.metadata:
            root.append(self.visit(schematron_output.metadata))

        for validation_event in schematron_output.validation_events:
            el = self.visit(validation_event)
            if el is not None:
                root.append(el)

        return root

    def _process_ns_prefix_node(self, ns_prefix: NSPrefixInAttributeValues) -> _Element:
        """Process the namespace prefix node.

        Args:
            ns_prefix: the prefix node to render into an XML element

        Returns:
            The created XML element
        """
        return Element(f'{{{self._nsmap["svrl"]}}}ns-prefix-in-attribute-values',
                       attrib={'prefix': ns_prefix.prefix, 'uri': ns_prefix.uri},
                       nsmap=self._nsmap)

    def _process_metadata(self, metadata: MetaData) -> _Element:
        """Process the metadata node.

        Args:
            metadata: the metadata node

        Returns:
            The created XML node
        """
        additional_namespaces = {}
        for namespace in metadata.namespaces:
            additional_namespaces[namespace.prefix] = namespace.uri

        metadata_root = Element(f'{{{self._nsmap["svrl"]}}}metadata', nsmap=self._nsmap | additional_namespaces)
        for element in metadata.xml_elements:
            metadata_root.append(element)

        return metadata_root

    def _process_active_pattern_node(self, active_pattern: ActivePattern) -> _Element:
        """Process an active pattern node.

        Args:
            active_pattern: The active pattern node

        Returns:
            The created XML element
        """
        documents = None
        if active_pattern.documents:
            documents = ' '.join([f'file:{doc}' for doc in active_pattern.documents])

        node_attributes = {
            'documents': documents,
            'id': active_pattern.id,
            'name': active_pattern.name,
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        return Element(f'{{{self._nsmap["svrl"]}}}active-pattern', attrib=final_attributes, nsmap=self._nsmap)

    def _process_fired_rule_node(self, fired_rule: FiredRule) -> _Element:
        """Process a fired rule node.

        Args:
            fired_rule: information on the fired rule

        Returns:
            The created XML element
        """
        document = None
        if fired_rule.document:
            document = f'file:{fired_rule.document}'

        node_attributes = {
            'context': fired_rule.context.query,
            'document': document,
            'flag': fired_rule.flag,
            'id': fired_rule.id,
            'name': fired_rule.name,
            'role': fired_rule.role
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        return Element(f'{{{self._nsmap["svrl"]}}}fired-rule', attrib=final_attributes, nsmap=self._nsmap)

    def _process_suppressed_rule_node(self, suppressed_rule: SuppressedRule) -> _Element:
        """Process a suppressed rule node

        Args:
            suppressed_rule: information on the suppressed rule

        Returns:
            The created XML element
        """
        node_attributes = {
            'context': suppressed_rule.context.query,
            'id': suppressed_rule.id,
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        return Element(f'{{{self._nsmap["svrl"]}}}suppressed-rule', attrib=final_attributes, nsmap=self._nsmap)

    def _process_failed_assert(self, failed_assert: FailedAssert) -> _Element:
        """Process a failed assert node.

        Args:
            failed_assert: the failed assert information

        Returns:
            The created XML element
        """
        return self._process_check_result(failed_assert)

    def _process_successful_report(self, successful_report: SuccessfulReport) -> _Element:
        """Process a successful report node.

        Args:
            successful_report: the successful report information

        Returns:
            The created XML element
        """
        return self._process_check_result(successful_report)

    def _process_check_result(self, check_result: FailedAssert | SuccessfulReport):
        """Process a check result.

        Depending on the type of input we return either a `successful-report` or a `failed-assert` element.

        Args:
            check_result: the check result.

        Returns:
            The created XML element
        """
        node_attributes = {
            'flag': check_result.flag,
            'id': check_result.id,
            'location': check_result.location.expression,
            'role': check_result.role,
            'test': check_result.test.query
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        if subject_location := check_result.subject_location:
            final_attributes['location'] = subject_location.expression

        node_name = 'successful-report'
        if isinstance(check_result, FailedAssert):
            node_name = 'failed-assert'

        report_element = Element(f'{{{self._nsmap["svrl"]}}}{node_name}',
                                 attrib=final_attributes, nsmap=self._nsmap)

        if check_result.diagnostic_references:
            for diagnostic_reference in check_result.diagnostic_references:
                report_element.append(self.visit(diagnostic_reference))

        if check_result.property_references:
            for property_reference in check_result.property_references:
                report_element.append(self.visit(property_reference))

        if check_result.text.content:
            report_element.append(self.visit(check_result.text))

        return report_element

    def _process_text_node(self, text: Text) -> _Element:
        """Process a text node.

        Args:
            text: the text node to convert into an element.

        Returns:
            The created XML element
        """
        node_attributes = {
            'fpi': text.fpi,
            'icon': text.icon,
            'see': text.see,
            'class': text.class_,
            'id': text.id,
            '{http://www.w3.org/XML/1998/namespace}lang': text.xml_lang,
            '{http://www.w3.org/XML/1998/namespace}space': text.xml_space,
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        text_element = Element(f'{{{self._nsmap["svrl"]}}}text', attrib=final_attributes, nsmap=self._nsmap)
        text_element.text = text.content
        return text_element

    def _process_property_reference(self, property_reference: PropertyReference) -> _Element:
        """Process a property reference node.

        Args:
            property_reference: the property reference to convert into a node.

        Returns:
            The created XML element
        """
        node_attributes = {
            'property': property_reference.property,
            'role': property_reference.role,
            'scheme': property_reference.scheme
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        property_reference_element = Element(f'{{{self._nsmap["svrl"]}}}property-reference',
                                             attrib=final_attributes, nsmap=self._nsmap)

        if property_reference.text:
            property_reference_element.append(self.visit(property_reference.text))

        return property_reference_element

    def _process_diagnostic_reference(self, diagnostic_reference: DiagnosticReference) -> _Element:
        """Process a diagnostic reference node.

        Args:
            diagnostic_reference: the diagnostic reference to convert into a node.

        Returns:
            The created XML element
        """
        node_attributes = {
            'diagnostic': diagnostic_reference.diagnostic,
        }
        final_attributes = {k: v for k, v in node_attributes.items() if v is not None}

        diagnostic_reference_element = Element(f'{{{self._nsmap["svrl"]}}}diagnostic-reference',
                                               attrib=final_attributes, nsmap=self._nsmap)

        if diagnostic_reference.text:
            diagnostic_reference_element.append(self.visit(diagnostic_reference.text))

        return diagnostic_reference_element



"""Definition of the common API for validating an XML using Schematron.

This defines a common interface for Schematron validators to implement.

At the moment this is only implemented by direct mode evaluation, but in the future it might also support
the XSLT evaluation.
"""
from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2024-04-01'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from abc import ABCMeta, abstractmethod
from pathlib import Path

from lxml.etree import _ElementTree

from pyschematron.direct_mode.xml_validation.queries.base import CustomQueryFunction


class SchematronValidatorFactory(metaclass=ABCMeta):
    """Factory class for generating Schematron validators.

    The Schematron validators do the hard work of validating an XML document. To ensure that a single validator can
    process multiple XML documents once set on a phase, we make the validators immutable and built them using this
    factory.
    """

    @abstractmethod
    def set_schema(self, schematron_xml: Path | _ElementTree):
        """Set the Schematron schema.

        Args:
            schematron_xml: the Schematron Schema we would like to use in the validation.
        """

    @abstractmethod
    def set_base_path(self, schematron_base_path: Path):
        """Set the Schematron base path.

        Some Schematron schemas include files from other locations. If you would like explicit control over the
        Schematron base path, or if the path could not be inferred from the provided schema, set it here explicitly.

        Args:
            schematron_base_path: the base path for the Schematron definition
        """

    @abstractmethod
    def set_phase(self, phase: str | None):
        """Set the phase we would like the Schematron validator to validate.

        By setting this before we load the Schematron validator, we can prepare the phase for evaluation, giving
        us some speed benefit.

        Args:
            phase: the phase we would like to validate. If set to None we use the '#DEFAULT' phase.
        """

    @abstractmethod
    def build(self) -> SchematronValidator:
        """Construct the configured Schematron validator.

        Returns:
            An implementation of the Schematron validator.
        """

    @abstractmethod
    def add_custom_functions(self,
                             query_binding: str,
                             custom_query_functions: list[CustomQueryFunction],
                             base_query_binding: str | None = None):
        """Add custom functions to the parser we would like to use.

        The custom query functions are added to the parser for a specific query binding language. Repeated calls
        accumulate the additional functions. If you create a new query binding specifier, you need to specify the base
        query binding language to use as basis for the new query binding language.

        Args:
            query_binding: the query binding language for which we are adding the custom query functions
            custom_query_functions: a list of custom query function objects
            base_query_binding: the basis to use for any new query binding language
        """


class SchematronValidator(metaclass=ABCMeta):
    """The Schematron validator, validating an XML document using the init injected Schematron definition."""

    @abstractmethod
    def validate(self, xml_data: Path | _ElementTree) -> ValidationResult:
        """Validate an XML document and return an SVRL XML document.

        Args:
            xml_data: the XML data we would like to validate

        Returns:
            The validation results.
        """


class ValidationResult(metaclass=ABCMeta):
    """Results from Schematron validation.

    This should be able to produce an SVRL XML document, and should be able to tell if the document was valid or not.
    """

    @abstractmethod
    def get_svrl(self) -> _ElementTree:
        """Get the SVRL as an XML element tree.

        Returns:
            The SVRL as an element tree.
        """

    @abstractmethod
    def is_valid(self) -> bool:
        """Check if the document we validated was valid.

        According to the specifications, a successful report is considered a failure. As such, this method considers
        an XML document to be valid if none of the assertions and none of the reports were raised.

        Returns:
            If the document was valid return True, else False.
        """
